//Simplified BSD License (BSD-2-Clause)
//
//Copyright (c) 2024, Aditiya Afrizal, The OpenBVE Project
//
//Redistribution and use in source and binary forms, with or without
//modification, are permitted provided that the following conditions are met:
//
//1. Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
//2. Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
//
//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
//ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#version 410 core

in vec2 vUv;

uniform sampler2D uTexture;
uniform bool uHasTexture;
uniform float uAlphaCutoff;
uniform float uMaterialAlpha; // Material color alpha (0.0–1.0), allows semi-transparent faces to skip shadow casting
uniform int uMaterialFlags;

out vec4 fragColor;

void main()
{
    // Start with the base material alpha
    float alpha = uMaterialAlpha;

    if (uHasTexture)
    {
        alpha *= texture(uTexture, vUv).a;
    }

    // Discard fragments whose effective alpha is below the cutoff
    // This prevents semi-transparent glass meshes from casting full opaque shadows
	// Also ignore materials with no shadow flag set (ground / crack)
    if (alpha <= uAlphaCutoff || ((uMaterialFlags & 32) != 0))
    {
        discard;
    }

    fragColor = vec4(1.0);
}

